"""Dataclasses, enums, and  types related to Tamr"""
from . import project_type
from . import data_type
from . import operation_state
from . import validation_check

__all__ = ["project_type", "data_type", "operation_state", "validation_check"]



"""Tasks related to the version of Tamr instances"""
from tamr_unify_client import Client
from typing import Callable
import json
import inspect

import logging

LOGGER = logging.getLogger(__name__)


def current(client: Client) -> str:
    """Gets the version of Tamr for provided client

    Args:
        client: Tamr client

    Returns:
        String representation of Tamr version

    """
    url = "/api/versioned/service/version"
    response = client.get(url).successful()
    return json.loads(response.content)["version"]


def _as_float(version: str) -> float:
    """Converts string Tamr version to an orderable numeric representation

    The numeric representation is designed to allow for the ordering of Tamr versions from oldest
    to newest. The values are not guaranteed to be sequential even if the versions are sequential.
    Formats covered:
        year.release.patch (2020.001.1)
        major.minor.patch (0.43.0)

    Args:
        version: String representation of Tamr version

    Returns:
        Numeric representation of Tamr version

    """
    version_split = version.split(".")

    if len(version_split) != 3:
        raise ValueError(f"Tamr version {version} does not match known patterns.")

    version_split = [float(x) for x in version_split]
    return (version_split[0] * 1000) + version_split[1] + (version_split[2] / 10)


def enforce_after_or_equal(client: Client, *, compare_version: str) -> None:
    """Raises an exception if the version of the Tamr client is before the provided compare version

    Args:
        client: Tamr client
        compare_version: String representation of Tamr version

    Returns:
        None

    """
    current_version = current(client)
    if _as_float(current_version) < _as_float(compare_version):
        raise NotImplementedError(
            f"This function is not available in Tamr {current_version}. "
            f"Upgrade to Tamr {compare_version} or later to use this function."
        )


def _deprecated_warning(func: Callable, *, message: str) -> Callable:
    """Decorator to log a warning message when the passed function is called.
    Intended for warning about deprecated functions.

    Args:
        func: The function to attach the warning message to
        message: The warning message
    Returns:
        The decorated function
    """

    def warning(*args, **kwargs):
        try:
            current_frame = inspect.currentframe()
            previous_frame = current_frame.f_back
            calling_lineno = previous_frame.f_lineno
            calling_func = previous_frame.f_code
            LOGGER.warning(
                f"In {calling_func.co_filename}:{calling_func.co_name}:{calling_lineno} {message}"
            )
        finally:
            # Avoid reference cycle
            # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
            del current_frame
            del previous_frame
        return func(*args, **kwargs)

    return warning



"""Tasks related to Tamr Projects"""
from . import schema_mapping
from . import mastering
from . import golden_records
from . import categorization

__all__ = ["schema_mapping", "mastering", "golden_records", "categorization"]



"""Tasks related to Tamr operations (or jobs)"""
import logging
from typing import Union, List

from tamr_unify_client import Client
from tamr_unify_client.operation import Operation
from tamr_toolbox.models.operation_state import OperationState

LOGGER = logging.getLogger(__name__)


def enforce_success(operation: Operation) -> None:
    """Raises an error if an operation fails

    Args:
        operation: A Tamr operation
    """
    if not operation.succeeded():
        raise RuntimeError(
            f"Operation {operation.resource_id} failed. Description: {operation.description}."
            f"Status: {operation.status}"
        )


def from_resource_id(tamr: Client, *, job_id: Union[int, str]) -> Operation:
    """Create an operation from a job id

    Args:
        tamr: A Tamr client
        job_id: A job ID

    Returns:
        A Tamr operation
    """
    return Operation.from_resource_id(tamr, str(job_id))


def get_latest(tamr: Client) -> Operation:
    """ Get the latest operation

    Args:
        tamr: A Tamr client

    Returns:
        The latest job
    """
    op = get_all(tamr)[0]
    return op


def get_details(*, operation: Operation) -> str:
    """Return a text describing the information of a job

    Args:
        operation: A Tamr operation

    Returns:
        A text describing the information of a job
    """
    return (
        f"Host: {operation.client.host} \n Job: {operation.resource_id} \n "
        f"Description: {operation.description} \n Status: {operation.state} "
    )


def get_all(tamr: Client) -> List[Operation]:
    """
    Get a list of all jobs or operations.

    Args:
        tamr: A Tamr client

    Returns:
        A list of Operation objects.

    """
    response = tamr.get(
        "/api/versioned/v1/operations", headers={"Accept": "application/json"}, stream=True
    ).json()
    ops = [Operation.from_json(tamr, item) for item in response]

    return ops


def get_active(tamr: Client) -> List[Operation]:
    """
    Get a list of pending and running jobs.

    Args:
        tamr: A Tamr client

    Returns:
        A list of Operations objects

    """
    ops = get_all(tamr)
    active_states = [OperationState.PENDING, OperationState.RUNNING]
    active_ops = [op for op in ops if OperationState(op.state) in active_states]

    return active_ops



"""Tasks common to moving data in and out of Tamr"""
from typing import Optional, List, Any, Iterable, Callable
import logging

from tamr_unify_client.dataset.resource import Dataset

LOGGER = logging.getLogger(__name__)


def _flatten_list(val: Any, *, delimiter: str = "|", force: bool = False) -> Any:
    """
    Flattens list types to strings.
    By default, only list[string] is flattened to a single string.
    The `force` parameter allows flattening of other inner types like numerics to a string.
    A list of length 1 becomes a string.
    A list with length longer than 1 has its elements concatenated with the specified `delimiter`.
    Non-list types are passed through directly.
    Intended to be applied using `applymap` to a DataFrame created from a Tamr Dataset,
    or to a pandas Series using `apply`.

    Args:
        val: value to convert
        delimiter: string to use as delimiter for concatenating lists
        force: if True, will force non-string inner types to string

    Returns:
        string or original value
    """
    if isinstance(val, list):
        if force:
            # convert inner values to strings.  f-strings are faster than str()
            return delimiter.join(f"{i}" for i in filter(None, val))
        else:
            return delimiter.join(filter(None, val))
    else:
        return val


def _yield_records(
    dataset: Dataset,
    *,
    func: Optional[Callable] = None,
    columns: Optional[List] = None,
    flatten_columns: Optional[List] = None,
) -> Iterable:
    """
    Generator function for records from a Tamr Dataset.
    Can optionally apply a flattening function to some or all columns,
    and can optionally keep only a subset of columns from the original records.

    Args:
        dataset: Tamr Dataset
        func: optional, callable function to transform records
        columns: optional, list of columns to keep
        flatten_columns: optional, list of columns to flatten

    Returns: iterable over Dataset records

    Raises:
        ValueError: if `flatten_columns` is not None but `func` is None since it
            won't know how to flatten them.
    """

    if flatten_columns is not None and func is None:
        message = (
            f"Columns specified for flatting ('flatten_columns'={flatten_columns}), "
            f"but no flattening function provided for parameter 'func'"
        )
        LOGGER.error(message)
        raise ValueError(message)

    checked_columns = False
    for record in dataset.records():
        if not checked_columns:
            if columns is not None:
                _check_columns_subset(
                    input_list=columns, reference_list=record.keys(), raise_error=True
                )
            if flatten_columns is not None:
                _check_columns_subset(
                    input_list=flatten_columns, reference_list=record.keys(), raise_error=True
                )
            checked_columns = True

        # Set flatten_columns to all if unspecified
        if func is not None and flatten_columns is None:
            flatten_columns = record.keys()

        output = {}
        for k, v in record.items():
            if columns is not None and k not in columns:
                # remove the column by skipping
                continue

            do_flatten_column = flatten_columns is not None and k in flatten_columns
            if do_flatten_column and func is not None:
                output[k] = func(v)
            else:
                output[k] = v
        yield output


def _check_columns_subset(
    *, input_list: List, reference_list: List, raise_error: bool = True
) -> bool:
    """
    Checks if `input_list` is a subset of `reference_list`.
    Returns a bool, or if `raise_error` is True, raises a ValueError on a failed check.

    Args:
        input_list: list to perform check for
        reference_list: reference list to check against for membership
        raise_error: if True, will raise a ValueError on failed check instead of returning

    Returns: bool

    Raises:
        ValueError: if `input_list` is not a subset of `reference_list`, and `raise_error` is True
    """
    input_set = set(input_list)
    reference_set = set(reference_list)
    if not input_set.issubset(reference_set):
        if raise_error:
            message = f"column(s) {input_set.difference(reference_set)} not found"
            LOGGER.error(message)
            raise ValueError(message)
        else:
            return False
    return True



"""Tasks related to running jobs for Tamr Schema Mapping projects"""
from typing import List
import logging

from tamr_unify_client.operation import Operation
from tamr_unify_client.project.resource import Project

from tamr_toolbox.models.project_type import ProjectType
from tamr_toolbox.utils import operation

LOGGER = logging.getLogger(__name__)


def _run_custom(project: Project, *, run_update_unified_dataset=False) -> List[Operation]:
    """Executes specified steps of a schema mapping project.

    Args:
        project: Target schema mapping project
        run_update_unified_dataset: Whether refresh should be called on the unified dataset

    Returns:
        The operations that were run

    Raises:
        TypeError: if the `project` is not a Schema Mapping project
    """
    if ProjectType[project.type] != ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS:
        error_msg = f"Cannot use as a schema mapping project. Project type: {project.type}"
        LOGGER.error(error_msg)
        raise TypeError(error_msg)

    completed_operations = []
    if run_update_unified_dataset:
        LOGGER.info(
            f"Updating the unified dataset for project {project.name} (id={project.resource_id})."
        )
        op = project.unified_dataset().refresh()
        operation.enforce_success(op)
        completed_operations.append(op)

    return completed_operations


def run(project: Project) -> List[Operation]:
    """Run the project

    Args:
        project: Target schema mapping project

    Returns:
        The operations that were run
    """
    return _run_custom(project, run_update_unified_dataset=True)


def update_unified_dataset(project: Project) -> List[Operation]:
    """Updates the unified dataset for a schema mapping project

    Args:
        project: Target schema mapping project

    Returns:
        The operations that were run
    """
    return _run_custom(project, run_update_unified_dataset=True)



"""Tasks related to creation of Slack notifications"""
import logging
import time
import os
from typing import Union, List, Optional

from tamr_unify_client import Client
from tamr_unify_client.operation import Operation

from tamr_toolbox.models.operation_state import OperationState
from tamr_toolbox.utils.operation import get_details, from_resource_id

LOGGER = logging.getLogger(__name__)

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import slack


def send_message(
    *, slack_client: "slack.WebClient", channel: str, message: str, raise_error: bool = True,
) -> dict:
    """Sends a message to a pre-defined Slack channel

    Args:
        slack_client: A Slack WebClient
        channel: A (public) Slack channel to receive messages
        message: A message to be sent to a Slack channel
        raise_error: A boolean value to opt out raising Slack API errors

    Returns:
        The data property of the Slack WebClient object

    Raises:
        SlackApiError: if the SlackClient does not post the message correctly
            and `raise_error` is set to True
    """

    # This function requires slack, an optional dependency
    from slack.errors import SlackApiError

    try:
        slack_response = slack_client.chat_postMessage(channel=channel, text=message)
    except SlackApiError as e:
        LOGGER.error(f"Error: {e}")

        if not raise_error:
            return {
                "type": "SlackError",
                "text": f'The message: "{message}" failed to send to channel: {channel}',
                "details": e.response["error"],
            }
        else:
            raise e

    else:
        return slack_response.data


def _send_job_status_message(
    *,
    slack_client: "slack.WebClient",
    channel: str,
    operation: Operation,
    notify_states: List[OperationState],
) -> dict:
    """
    Checks operation state and if in `notify_states` sends the message.

    Args:
        slack_client: A Slack WebClient
        channel: A (public) Slack channel to receive messages
        operation: A Tamr Operation
        notify_states: States for which notifications should be sent

    Returns:
        The data property of the Slack WebClient object
    """
    state = OperationState[operation.state]
    if state in notify_states:
        message = get_details(operation=operation)
        return send_message(slack_client=slack_client, channel=channel, message=message)


def monitor_job(
    tamr: Client,
    *,
    slack_client: "slack.WebClient",
    channel: str,
    operation: Union[int, str, Operation],
    poll_interval_seconds: float = 1,
    timeout_seconds: Optional[float] = None,
    notify_states: Optional[List[OperationState]] = None,
) -> List[dict]:
    """

    Monitors a Tamr Operation and sends a Slack message to a channel when the job status is updated

    Args:
        tamr: A Tamr client
        slack_client: A Slack WebClient
        channel: A (public) Slack channel to receive messages
        operation: A job ID or a Tamr operation
        poll_interval_seconds: Time interval (in seconds) between subsequent polls
        timeout_seconds: Time (in seconds) to wait
        notify_states : States for which notifications should be sent, use None for all states

    Returns:
        A list of messages with their config sent via the Slack WebClient object
    """
    if notify_states is None:
        notify_states = [
            OperationState.SUCCEEDED,
            OperationState.FAILED,
            OperationState.CANCELED,
            OperationState.PENDING,
            OperationState.RUNNING,
        ]
    started = time.time()
    list_responses = []
    status = None

    if isinstance(operation, Operation):
        op = operation
    else:
        op = from_resource_id(tamr=tamr, job_id=operation)

    while (timeout_seconds is None or time.time() - started < timeout_seconds) and status not in [
        OperationState.SUCCEEDED,
        OperationState.FAILED,
        OperationState.CANCELED,
    ]:
        # Check the status of the current operation.
        # If the state is updated such that status!=new_status then send a message and
        # update `status`.
        # The loop exits if hits the timeout or if the jobs is in
        # a final state (SUCCEEDED/FAILED/CANCELED).
        # Note that it will always send a message about any status that has been
        # updated before exiting
        op = op.poll()
        new_status = OperationState[op.state]
        if status != new_status:
            message = _send_job_status_message(
                slack_client=slack_client,
                channel=channel,
                operation=op,
                notify_states=notify_states,
            )
            list_responses.append(message)
            status = new_status
        time.sleep(poll_interval_seconds)

    if status not in [
        OperationState.SUCCEEDED,
        OperationState.FAILED,
        OperationState.CANCELED,
    ]:
        # If the operation was not in a final state then assume it timed out
        message = send_message(
            slack_client=slack_client,
            channel=channel,
            message=(
                f"The job {op.resource_id}: {op.description} took longer "
                f"than {timeout_seconds} seconds to resolve."
            ),
        )
        list_responses.append(message)

    return list_responses
